// microtuner.c - (C) Victor Cerullo 2007
// version 1.1 - April 2007
//
// microtuner is an enhancement of the mtx_loader object
// version 1.1 is a multiplatform code version compatible with Apple Xcode and Cygwin gcc for Windows

#include <ext.h>
#include <ext_path.h>
#include <math.h>

#define FTableSize 256  // double the keymap size to allow bend and modulation up to 64 half-steps
#define FTableOffset 64 // offsets to the actual MIDI note keymap [64..191]

typedef struct microtuner
{
    t_object p_ob;
    long ModRange[3]; // bend/modulation ranges (for outlet couples r1, r2, r3)
    long p_value0;    // MIDI note value - received from left inlet
    void *p_outlet;   // frequency float outlet
    void *r1_min;     // lower limit in cent (modulation control 1)
    void *r1_max;     // higher limit in cent (modulation control 1)
    void *r2_min;     // lower limit in cent (modulation control 2)
    void *r2_max;     // higher limit in cent (modulation control 2)
    void *r3_min;     // lower limit in cent (modulation control 3)
    void *r3_max;     // higher limit in cent (modulation control 3)
    void *p_out1;     // list of MIDI cents (for sending to a coll object)
    void *p_out2;     // list of frequencies (for sending to a coll object)
    void *p_out3;     // rightmost outlet bangs in case of successful scale expansion
} t_microtuner;

double FTable[FTableSize]; // Frequency table
t_symbol *ps_nothing;
void *microtuner_class;
void microtuner_expand(t_microtuner *x, short key, short period, short intervals);
void microtuner_read(t_microtuner *x, t_symbol *s);
void microtuner_doread(t_microtuner *x, t_symbol *s, short argc, t_atom *argv);
void microtuner_reset(t_microtuner *x);
void microtuner_dump(t_microtuner *x);
void microtuner_sendlist(t_microtuner *x);
void microtuner_dumpall(t_microtuner *x);
void microtuner_sendall(t_microtuner *x);
void microtuner_loadbang(t_microtuner *x);
void microtuner_int(t_microtuner *x, long n);
void microtuner_list(t_microtuner *x, t_symbol *s, short ac, t_atom *av);
void microtuner_setmidi(t_microtuner *x, t_symbol *s, short ac, t_atom *av);
void microtuner_range(t_microtuner *x, t_symbol *s, short ac, t_atom *av);
void microtuner_et(t_microtuner *x, t_symbol *s, short ac, t_atom *av);
void microtuner_retune(t_microtuner *x, t_symbol *s, short ac, t_atom *av);
void microtuner_cent(t_microtuner *x, t_symbol *s, short ac, t_atom *av);
void microtuner_assist(t_microtuner *x, void *b, long m, long a, char *s);
void *microtuner_new(long n);

void main(void)
{
    setup((t_messlist **)&microtuner_class, (method)microtuner_new, 0L, (short)sizeof(t_microtuner), 0L, A_DEFLONG, 0);
    addint((method)microtuner_int);
    addmess((method)microtuner_read, "read", A_DEFSYM,0);         // open, read and process tuning file
    addmess((method)microtuner_reset, "reset", A_DEFSYM,0);       // reset to 12-tET keymap
    addmess((method)microtuner_dump, "dump", A_DEFSYM,0);         // bang and dump MIDI cent and MIDI frequency keymap [0..127]
    addmess((method)microtuner_dumpall, "dumpall", A_DEFSYM,0);   // bang and dump full internal frequency table [0..FTableSize]
    addmess((method)microtuner_list,"list", A_GIMME,0);           // direct assignment of frequency to FTable cell n [0..FTableSize]
    addmess((method)microtuner_setmidi,"setmidi", A_GIMME,0);     // direct assignment of frequency to MIDI note n [0..127]
    addmess((method)microtuner_range,"range", A_GIMME,0);         // set bend/modulation range (ranges 1, 2 and 3) [0..64]
    addmess((method)microtuner_et,"et", A_GIMME,0);               // create equal tempered scale
    addmess((method)microtuner_retune,"retune", A_GIMME,0);       // retune keymap to a given reference frequency (MIDI note n, frequency)
    addmess((method)microtuner_cent,"cent", A_GIMME,0);           // transpose keymap (+/- cent)
    addmess((method)microtuner_assist, "assist", A_CANT, 0);
    addmess((method)microtuner_loadbang, "loadbang", A_CANT, 0);
    ps_nothing = gensym("");
}

void microtuner_expand(t_microtuner *x, short key, short period, short intervals)
{
    double FRatio[FTableSize];
    short i, count, OffsetKey;
    short UpperLimit;
    OffsetKey = key + FTableOffset;
    if (intervals==0) {
        // case of ConsiderIntervals checkbox not marked
        for (i=OffsetKey-1; i >= 0; --i) FTable[i]=FTable[i+period]/2;
        for (i=OffsetKey+period; i < FTableSize; ++i) FTable[i]=FTable[i-period]*2;
        microtuner_dump(x);
    }
    else {
        // case of ConsiderIntervals checkbox marked
        if (OffsetKey+period-1 > (FTableSize-2)) UpperLimit=(FTableSize-2); else UpperLimit=OffsetKey+period-1;
        for (i=OffsetKey; i<=UpperLimit; ++i) {
            if (FTable[i]!=0) FRatio[i-OffsetKey]=FTable[i+1]/(FTable[i]); else {
                if (FTable[i+1]==0) FRatio[i-OffsetKey]=1; else FRatio[i-OffsetKey]=999999999;
        }}
        if (period>1) {
            count=0;
            for (i=OffsetKey+period; i<=(FTableSize-1); ++i) {
                FTable[i]=FTable[i-1]*FRatio[count];
                ++count;
                if (count>period-2) count=0;
            }
            count=period-2;
            for (i=1; i<=OffsetKey; ++i) {
                FTable[OffsetKey-i]=FTable[OffsetKey-i+1]/FRatio[count];
                --count;
                if (count<0) count=period-2;
            }
            microtuner_dump(x);
        }
        else {
            post("microtuner error: INTERVALS mode requires at least two tones");
    }}
}

void microtuner_read(t_microtuner *x, t_symbol *s) {
    defer(x,(method)microtuner_doread,s,0,0);
}

void microtuner_doread(t_microtuner *x, t_symbol *s, short argc, t_atom *argv)
{
    char filename[256], fname[256];
    filename[0] = 0;
    fname[0] = 0;
    t_filehandle f_fh;
    long size, bytecount;
    bytecount = 1;
    Byte data[16];
    short i, path, NoGoodFile;
    long outtype;
    const int MaxLineSize = 1024;
    char line[MaxLineSize];  // Single line read from file
    char c;
    short pos, keyline, freqline;
    int key;             // Scale start MIDI note
    int period;          // Number of tones the scale is composed of
    short intervals;     // "Consider Intervals" flag
    if (s==ps_nothing) {
        open_promptset("Select a Microtuner tuning file (.mtx)");
        if (open_dialog(filename, &path, &outtype, 0L, 0)) {
            post("microtuner warning: no file selected");
            return;
        }
        } else {
        strcpy(fname, s->s_name);
        strcpy(filename, s->s_name);
        if (locatefile_extended(fname, &path, &outtype, 0L, 0)) {
            post("microtuner error: file not found");
            return;
        }
    }
    if (!path_opensysfile(filename,path,&f_fh,READ_PERM)) {
        post("microtuner: parsing file %s", filename);
        intervals = 1;
        period = 0;
        key = 0;
        keyline = 0;
        freqline = 0;
        sysfile_geteof(f_fh,&size);
        NoGoodFile = 0;
        do {
            pos = 0;
            do {
                c = '\0';
                if (sysfile_read(f_fh,&bytecount,&data)) {
                    c = EOF;
                    break;
                }
                --size;
                c = (char)data[0];
                if ((c=='\0')||(c=='\r')||(c=='\n')||(c==EOF)||(size==0)) break;
                if ((pos==0)&&(keyline!=1)&&(isdigit(c)||(c=='.'))) freqline = 1;
                if ((pos==0)&&(c=='@')&&(keyline==0)) keyline = 1; else line[pos++] = c;
            } while ((c != EOF) && (size > 0) && (pos < MaxLineSize));
            if (pos < MaxLineSize) {
                line[pos] = '\0';
                if (strcmp(line,":absolute")==0) intervals=0;
                if (keyline==1) {
                    key = atoi(line);
                    if ((key>=0) && (key<128)){
                        post("microtuner: scale starts at MIDI key %d", key);
                        for (i=0; i < FTableSize; ++i) FTable[i]=0;
                        keyline = 99;
                    } else NoGoodFile = 1;
                }
                if ((keyline==99)&&(freqline==1)) {
                    if ((key+period) < 128) {
                        FTable[key+period+FTableOffset] = atof(line);
                        ++period;
                        freqline = 0;
                    } else NoGoodFile = 1;
                }
            } else NoGoodFile = 1;
        } while ((c != EOF) && (size > 0) && (!NoGoodFile));
        sysfile_close(f_fh);
        if ((keyline!=99)||(period==0)) NoGoodFile = 1;
        if (!NoGoodFile) {
            post("microtuner: number of tones in the scale = %d", period);
            if (!intervals) post("microtuner: expansion mode = ABSOLUTE");
            else post("microtuner: expansion mode = INTERVALS");
            microtuner_expand(x, key,period,intervals);
        } else post("microtuner error: file %s is not a valid mtx tuning file", &filename);
    }  else post("microtuner error: unable to open file %s", &filename);
}

void microtuner_reset(t_microtuner *x)
{
    // Initialize 12-tET default scale seed and expand keymap
    short i;
    for (i=0; i < FTableSize; ++i) FTable[i]=0;
    FTable[69+FTableOffset]=440;
    FTable[70+FTableOffset]=466.16376151809;
    microtuner_expand(x, 69,2,1);
}

void microtuner_dump(t_microtuner *x)
{
    outlet_bang (x->p_out3);
    microtuner_sendlist(x);
}

void microtuner_sendlist(t_microtuner *x)
{
    t_atom FreqList[2], CentList[2];
    short i;
    double ET;
    for (i=0; i < 128; ++i) {
        SETLONG(CentList, i);
        ET = 440 * pow(2, (((double)i-69)/12));
        SETFLOAT(CentList+1, (100*i+(log(FTable[i+FTableOffset])-log(ET))*1200/log(2)));
        SETLONG(FreqList, i);
        SETFLOAT(FreqList+1, FTable[i+FTableOffset]);
        outlet_list(x->p_out1, 0L, 2, CentList);
        outlet_list(x->p_out2, 0L, 2, FreqList);
    }
}

void microtuner_dumpall(t_microtuner *x)
{
    outlet_bang (x->p_out3);
    microtuner_sendall(x);
}

void microtuner_sendall(t_microtuner *x)
{
    t_atom FreqList[2], CentList[2];
    short i;
    double ET;
    for (i=0; i < FTableSize; ++i) {
        SETLONG(FreqList, i);
        SETFLOAT(FreqList+1, FTable[i]);
        outlet_list(x->p_out2, 0L, 2, FreqList);
        // also output MIDI cent list for the MIDI key subrange [64..191] (added in version 1.1)
        if ((i >= FTableOffset)&&(i < FTableOffset+128)) {
            SETLONG(CentList, i-FTableOffset);
            ET = 440 * pow(2, (((double)i-69-FTableOffset)/12));
            SETFLOAT(CentList+1, (100*(i-FTableOffset)+(log(FTable[i])-log(ET))*1200/log(2)));
            outlet_list(x->p_out1, 0L, 2, CentList);
        }
    }
}

void microtuner_int(t_microtuner *x, long n)
{
    double frequency, cent[6];
    short i, j, r;
    x->p_value0 = n;
    if (n<0) x->p_value0 = 0;
    if (n>127) x->p_value0 = 127;
    frequency = FTable[x->p_value0+FTableOffset];
    j = 0;
    for (i=0; i < 3; ++i) {
        r = x->ModRange[i];
        if (r<0) r = 0;
        if (r>FTableOffset) r = FTableOffset;
        cent[j++] = (log(FTable[x->p_value0+FTableOffset-r])-log(frequency))*1200/log(2);
        cent[j++] = (log(FTable[x->p_value0+FTableOffset+r])-log(frequency))*1200/log(2);
    }
    outlet_float(x->p_outlet, frequency);
    outlet_float(x->r1_min, cent[0]);
    outlet_float(x->r1_max, cent[1]);
    outlet_float(x->r2_min, cent[2]);
    outlet_float(x->r2_max, cent[3]);
    outlet_float(x->r3_min, cent[4]);
    outlet_float(x->r3_max, cent[5]);
}

void microtuner_list(t_microtuner *x, t_symbol *s, short ac, t_atom *av)
{
    short i;
    if (ac==2) {
        if (av->a_type==A_LONG) {
            if ((av->a_w.w_long >= 0)&&(av->a_w.w_long < FTableSize)) {
                i = av->a_w.w_long;
                av++;
                if (av->a_type==A_FLOAT) {
                    if (av->a_w.w_float >= 0) {
                        FTable[i]=(double)av->a_w.w_float;
                        microtuner_dump(x);
                    }
                }
                else if (av->a_type==A_LONG) {
                    if (av->a_w.w_long >= 0) {
                        FTable[i]=(double)av->a_w.w_long;
                        microtuner_dump(x);
                    }
                }
            }
        }
    }
}

void microtuner_setmidi(t_microtuner *x, t_symbol *s, short ac, t_atom *av)
{
    short i;
    if (ac==2) {
        if (av->a_type==A_LONG) {
            if ((av->a_w.w_long >= 0)&&(av->a_w.w_long < 128)) {
                i = av->a_w.w_long + FTableOffset;
                av++;
                if (av->a_type==A_FLOAT) {
                    if (av->a_w.w_float >= 0) {
                        FTable[i]=(double)av->a_w.w_float;
                        microtuner_dump(x);
                    }
                }
                else if (av->a_type==A_LONG) {
                    if (av->a_w.w_long >= 0) {
                        FTable[i]=(double)av->a_w.w_long;
                        microtuner_dump(x);
                    }
                }
            }
        }
    }
}

void microtuner_range(t_microtuner *x, t_symbol *s, short ac, t_atom *av)
{
    short i;
    if (ac==2) {
        if (av->a_type==A_LONG) {
            if ((av->a_w.w_long >= 1)&&(av->a_w.w_long <= 3)) {
                i = av->a_w.w_long;
                av++;
                if (av->a_type==A_LONG) {
                    if ((av->a_w.w_long >= 0)&&(av->a_w.w_long <= FTableOffset)) {
                        x->ModRange[i-1] = av->a_w.w_long;
                    }
                }
            }
        }
    }
}

void microtuner_et(t_microtuner *x, t_symbol *s, short ac, t_atom *av)
{
    short i, k;
    double j, ET, ET1;
    if (ac==3) {
        if (av->a_type==A_LONG) {
            if ((av->a_w.w_long >= 2)&&(av->a_w.w_long <= 128)) { // min 2 tones, max 128 tones
                i = av->a_w.w_long;
                av++;
                j = 0;
                if (av->a_type==A_LONG) {
                    if ((abs(av->a_w.w_long) > 1)&&(abs(av->a_w.w_long) <= 10)) { // width has to be greater than unison and <= 10
                        j = (double)av->a_w.w_long;
                    }
                }
                if (av->a_type==A_FLOAT) {
                    if ((abs(av->a_w.w_float) > 1)&&(abs(av->a_w.w_float) <= 10)) { // width has to be greater than unison and <= 10
                        j = (double)av->a_w.w_float;
                    }
                }
                av++;
                if ((j!=0)&&(av->a_type==A_LONG)) {
                    if ((av->a_w.w_long >= 0)&&(av->a_w.w_long < 128)) { // MIDI key has to be >= 0 and < 128
                        for (k=0; k < FTableSize; ++k) FTable[k]=0;
                        ET = 440 * pow(2, (((double)av->a_w.w_long-69)/12));
                        ET1 = ET * pow(j, 1/(double)i);
                        FTable[av->a_w.w_long+FTableOffset]=ET;
                        FTable[av->a_w.w_long+FTableOffset+1]=ET1;
                        microtuner_expand(x, av->a_w.w_long,2,TRUE);
                    }
                }
            }
        }
    }
}

void microtuner_retune(t_microtuner *x, t_symbol *s, short ac, t_atom *av)
{
    //retune only works when the frequency of the target MIDI note is > 0
    short i,j;
    double newref;
    if (ac==2) {
        if (av->a_type==A_LONG) {
            if ((av->a_w.w_long >= 0)&&(av->a_w.w_long < 128)) {
                i = av->a_w.w_long + FTableOffset;
                av++;
                if (av->a_type==A_FLOAT) {
                    newref = (double)av->a_w.w_float;
                    if ((newref > 0)&&(FTable[i] > 0)) {
                        for (j=0; j < FTableSize; ++j) FTable[j]=newref*(FTable[j]/FTable[i]);
                        microtuner_dump(x);
                    }
                }
                else if (av->a_type==A_LONG) {
                    newref = (double)av->a_w.w_long;
                    if ((newref > 0)&&(FTable[i] > 0)) {
                        for (j=0; j < FTableSize; ++j) FTable[j]=newref*(FTable[j]/FTable[i]);
                        microtuner_dump(x);
                    }
                }
            }
        }
    }
}

void microtuner_cent(t_microtuner *x, t_symbol *s, short ac, t_atom *av)
{
    short i;
    if (ac==1) {
        if (av->a_type==A_LONG) {
            for (i=0; i < FTableSize; ++i) FTable[i]=FTable[i]*pow(2, (double)av->a_w.w_long/1200);
            microtuner_dump(x);
        }
        else if (av->a_type==A_FLOAT) {
            for (i=0; i < FTableSize; ++i) FTable[i]=FTable[i]*pow(2, (double)av->a_w.w_float/1200);
            microtuner_dump(x);
        }
    }
}

void microtuner_assist(t_microtuner *x, void *b, long m, long a, char *s)
{
    if (m == ASSIST_INLET)
    sprintf(s,"MIDI note in");
    else {
        switch (a) {
            case 0:
            sprintf(s,"frequency out (Hz)");
            break;
            case 1:
            sprintf(s,"bend/mod range 1 min (Cent)");
            break;
            case 2:
            sprintf(s,"bend/mod range 1 max (Cent)");
            break;
            case 3:
            sprintf(s,"bend/mod range 2 min (Cent)");
            break;
            case 4:
            sprintf(s,"bend/mod range 2 max (Cent)");
            break;
            case 5:
            sprintf(s,"bend/mod range 3 min (Cent)");
            break;
            case 6:
            sprintf(s,"bend/mod range 3 max (Cent)");
            break;
            case 7:
            sprintf(s,"MIDI cent keymap dump (list)");
            break;
            case 8:
            sprintf(s,"frequency keymap dump (list)");
            break;
            case 9:
            sprintf(s,"bang after scale expansion");
            break;
        }
    }
}

void microtuner_loadbang(t_microtuner *x)
{
    microtuner_dump(x);
}

void *microtuner_new(long n)
{
    t_microtuner *x;
    x = (t_microtuner *)newobject(microtuner_class);
    x->p_out3 = outlet_new(x,0);
    x->p_out2 = listout(x);
    x->p_out1 = listout(x);
    x->r3_max = floatout(x);
    x->r3_min = floatout(x);
    x->r2_max = floatout(x);
    x->r2_min = floatout(x);
    x->r1_max = floatout(x);
    x->r1_min = floatout(x);
    x->p_outlet = floatout(x);
    x->p_value0    = 0;
    x->ModRange[0] = 2; // modulation default = 2 scale tones
    x->ModRange[1] = 2;
    x->ModRange[2] = 2;
    microtuner_reset(x); // always start in 12-tET mode
    return(x);
}
